@@ -1,9 +1,11 @@
+import time
 import json
 import logging
 from importlib import import_module
 from functools import wraps
 import inspect
 from os import walk, path
+import sys
 from urllib.parse import urlparse
 
 from asgiref.sync import async_to_sync
@@ -16,7 +18,7 @@ from django.conf import settings
 from .channel import Channel
 from .reflex import PROTECTED_VARIABLES
 from .element import Element
-from .utils import classify, camelize_value
+from .utils import classify
 
 
 logger = logging.getLogger('sockpuppet')
@@ -50,6 +52,9 @@ class SockpuppetConsumer(JsonWebsocketConsumer):
                 self.load_reflexes_from_config(config)
 
     def connect(self):
+        '''
+        We use the user session key as a default channel to publish any events
+        '''
         super().connect()
         session = self.scope['session']
         has_session_key = session.session_key
@@ -81,6 +86,9 @@ class SockpuppetConsumer(JsonWebsocketConsumer):
         )
 
     def disconnect(self, *args, **kwargs):
+        '''
+        When we disconnect we unsubscribe from the user session key.
+        '''
         session = self.scope['session']
         async_to_sync(self.channel_layer.group_discard)(
             session.session_key,
@@ -92,6 +100,10 @@ class SockpuppetConsumer(JsonWebsocketConsumer):
         )
         super().disconnect(*args, **kwargs)
 
+    def message(self, event):
+        logger.debug('Sending data: %s', event)
+        self.send(json.dumps(event))
+
     def group_send(self, recipient, message):
         send = async_to_sync(self.channel_layer.group_send)
         send(recipient, message)
@@ -126,14 +138,10 @@ class SockpuppetConsumer(JsonWebsocketConsumer):
 
                     append_reflex(module)
 
-    def message(self, event):
-        session_key = self.scope['session'].session_key
-        logger.debug('Sending data to session: %s, %s', session_key, event)
-        self.send(json.dumps(event))
-
-    def receive_json(self, data, **kwargs):
+    def reflex_message(self, data, **kwargs):
         logger.debug('Json: %s', data)
         logger.debug('kwargs: %s', kwargs)
+        start = time.perf_counter()
 
         url = data['url']
         selectors = data['selectors'] if data['selectors'] else ['body']
@@ -151,8 +159,10 @@ class SockpuppetConsumer(JsonWebsocketConsumer):
             msg = 'SockpuppetConsumer failed to invoke {target}, with url {url}, {message}'.format(
                 target=target, url=url, message=error
             )
-            self.broadcast_error(msg, data)
-            raise SockpuppetError(msg)
+            self.broadcast_error(msg, data, reflex)
+            _, _, traceback = sys.exc_info()
+            exc = SockpuppetError(msg)
+            raise exc.with_traceback(traceback)
 
         try:
             self.render_page_and_broadcast_morph(reflex, selectors, data)
@@ -161,13 +171,52 @@ class SockpuppetConsumer(JsonWebsocketConsumer):
             message = 'SockpuppetConsumer failed to re-render {url} {message}'.format(
                 url=url, message=error
             )
-            self.broadcast_error(message, data)
-            raise SockpuppetError(message)
+            self.broadcast_error(message, data, reflex)
+            _, _, traceback = sys.exc_info()
+            exc = SockpuppetError(msg)
+            raise exc.with_traceback(traceback)
+
+        logger.debug('Reflex took %6.2fms', (time.perf_counter() - start) * 1000)
+
+    def _get_channelname(self, channel_name):
+        try:
+            # StimulusReflex sends the channel name in the format
+            # of a json blob for name.
+            name = json.loads(channel_name)
+            name = name['channel'].replace('::', '-')
+        except json.decoder.JSONDecodeError:
+            name = channel_name
+        return name
+
+    def subscribe(self, data, **kwargs):
+        name = self._get_channelname(data['channelName'])
+        async_to_sync(self.channel_layer.group_add)(
+            name,
+            self.channel_name
+        )
+
+    def unsubscribe(self, data, **kwargs):
+        name = self._get_channelname(data['channelName'])
+        async_to_sync(self.channel_layer.group_discard)(
+            name,
+            self.channel_name
+        )
+
+    def receive_json(self, data, **kwargs):
+        message_type = data.get('type')
+        if message_type is None and data.get('target'):
+            self.reflex_message(data, **kwargs)
+        elif message_type == 'subscribe':
+            self.subscribe(data, **kwargs)
+        elif message_type == 'unsubscribe':
+            self.unsubscribe(data, **kwargs)
+        else:
+            print('Unsupported')
 
     def render_page_and_broadcast_morph(self, reflex, selectors, data):
         html = self.render_page(reflex)
         if html:
-            self.broadcast_morphs(selectors, data, html)
+            self.broadcast_morphs(selectors, data, html, reflex)
 
     def render_page(self, reflex):
         parsed_url = urlparse(reflex.url)
@@ -192,12 +241,12 @@ class SockpuppetConsumer(JsonWebsocketConsumer):
         reflex.session.save()
         return response.rendered_content
 
-    def broadcast_morphs(self, selectors, data, html):
+    def broadcast_morphs(self, selectors, data, html, reflex):
         document = BeautifulSoup(html)
         selectors = [selector for selector in selectors if document.select(selector)]
 
-        channel = Channel(self.scope['session'].session_key)
-        logger.debug('Broadcasting to %s', self.scope['session'].session_key)
+        channel = Channel(reflex.get_channel_id(), identifier=data['identifier'])
+        logger.debug('Broadcasting to %s', reflex.get_channel_id())
 
         for selector in selectors:
             channel.morph({
@@ -205,7 +254,7 @@ class SockpuppetConsumer(JsonWebsocketConsumer):
                 'html': ''.join([e.decode_contents() for e in document.select(selector)]),
                 'children_only': True,
                 'permanent_attribute_name': data['permanent_attribute_name'],
-                'stimulus_reflex': {**data, 'last': selector == selectors[-1]}
+                'stimulus_reflex': {**data}
             })
         channel.broadcast()
 
@@ -217,8 +266,10 @@ class SockpuppetConsumer(JsonWebsocketConsumer):
         else:
             getattr(reflex, method_name)(*arguments)
 
-    def broadcast_error(self, message, data):
-        channel = Channel(self.scope['session'].session_key)
+    def broadcast_error(self, message, data, reflex):
+        # We may have a sitation where we weren't able to get a reflex
+        session_key = reflex.get_channel_id() if reflex else self.scope['session'].session_key
+        channel = Channel(session_key)
         data.update({'error': 'message'})
         channel.dispatch_event({
             'name': 'stimulus-reflex:500',
